# coding:utf-8
# --------------------------------------------------------------------
# 名称:       ポリラインの Z値に放物線を設定 ツール(CreateParabolaLine tool)
# 作成日:     2020年5月10日
# Copyright:  WINGFIELD since1981
# 説明:       ライン フィーチャに指定したフィールドの値と係数値で放物線を描く
#             投影座標系のみサポート
# 操作方法:   下記URLを参照
#             https://www.wingfield.gr.jp/archives/11363
#
# 履歴:
#             2020-02-04  新規作成
# --------------------------------------------------------------------

import arcpy
import json
import sys
import os
import os.path
import distutils.util
import copy

try:
    ## 関数
    # 放物線を描く関数
    def calculate_parabola_y(x, y_top, x_max):
        # 放物線を描くための二次方程式
        # Y = c(X - a)^2 + b
        # a = 頂点のX座標
        # b = 頂点のY座標
        # c = 放物曲線の係数

        x_top = x_max / 2
        the_factor = -(y_top / x_top ** 2)
        return the_factor * (x - x_top) ** 2 + y_top


    ## 環境設定
    arcpy.env.outputZFlag = "Enabled"
    arcpy.env.outputMFlag = "Enabled"

    # テンポラリ フィールド名
    temp_field_name = "TEMP99999"
    temp_featureclass = r"memory\temp_output"

    # パラメーター
    input_features      = arcpy.GetParameterAsText(0)
    output_featureclass = arcpy.GetParameterAsText(1)
    factor_value        = arcpy.GetParameterAsText(2)
    factor_field        = arcpy.GetParameterAsText(3)

    # 入力フィーチャクラス
    if input_features == None or input_features == '':
        input_features = r"D:\\OneDrive\\GIS\\SpatialData\\Projects\\MyProject\\MyProject.gdb\\Polyline"

    # 出力フィーチャクラス
    if output_featureclass == None or output_featureclass == '':
        output_featureclass = r"D:\\OneDrive\\GIS\\SpatialData\\Projects\\MyProject\\MyProject.gdb\\Polyline_output"

    # 係数
    if factor_value == None or factor_value == '':
        factor_value = 1.0
    factor_value = float(factor_value)
    if factor_value == 0.0:
        arcpy.AddError('係数 0 は処理できません。')
        sys.exit()

    # 係数フィード
    if factor_field == None or factor_field == '':
        factor_field = ''

    # 頂点の挿入
    arcpy.management.GeodeticDensify(input_features, temp_featureclass, "GEODESIC")
    #arcpy.management.GeodeticDensify(input_features, output_featureclass, "GEODESIC") # M値に測地線距離を格納する場合は CreateRoutesが不要（未実装）

    # ルート フィールドの追加
    arcpy.management.AddField(temp_featureclass, temp_field_name, "LONG")

    # フィールド演算
    arcpy.management.CalculateField(temp_featureclass, temp_field_name,"!OBJECTID!")

    # M値の補間
    arcpy.lr.CreateRoutes(temp_featureclass, temp_field_name, output_featureclass)

    # フィーチャクラスの空間参照
    spatial_reference = arcpy.Describe(input_features).spatialReference

    # フィールド
    if factor_field == '':
        fields = ['SHAPE@JSON','OID@']
    else:
        fields = ['SHAPE@JSON', factor_field, 'OID@']

    # Z値の更新
    update_cursor = arcpy.da.UpdateCursor(output_featureclass, fields)
    
    for row in update_cursor:
        polyline = row[0]
        polyline_json = json.loads(polyline)
        
        # 放物線の頂点高の取得
        if factor_field == '':
            factor = factor_value
        else:
            factor = factor_value * row[1]

        # ポリラインが直線セグメント以外の場合は例外
        if polyline_json.get('paths') is None:
            print('入力フィーチャのジオメトリは直線セグメント以外をサポートしていません。')
            arcpy.AddError('入力フィーチャのジオメトリは直線セグメント以外をサポートしていません。')
            sys.exit()


        # M値に測地線距離を格納（未実装）
        # total_distance = 0.0
        # for i, value_i in enumerate(polyline_json['paths']):
        #     for j, value_j in enumerate(value_i):
        #         # セグメント間の測地線距離計算
        #         if i == 0 and j == 0:   # 最初の頂点
        #             polyline_json['paths'][i][j][3] = 0.0
        #         elif i > 0 and j == 0:  # マルチパートで次のセグメントの最初の頂点
        #             polyline_json['paths'][i][j][3] = copy.copy(polyline_json['paths'][i-1][-1][3])
        #         else:
        #             array = arcpy.Array([arcpy.Point(value_i[j-1][0], value_i[j-1][1]), arcpy.Point(value_i[j][0], value_i[j][0])])
        #             line = arcpy.Polyline(array, spatial_reference)
        #             total_distance += line.getLength('GEODESIC', 'METERS')
        #             polyline_json['paths'][i][j][3] = total_distance                   

        # M値の最大値（最後の頂点）を取得
        max_m = polyline_json['paths'][-1][-1][3]   #最終セグメント・最終頂点・M値

        # マルチパート ポリラインが生成された場合はパスの M値を逆転（経度180を跨ぐ場合の対処）
        if spatial_reference.type == 'Geographic' and len(polyline_json['paths']) > 1:
            copy_paths = copy.copy(polyline_json['paths'])
            for i, value_i in enumerate(copy_paths):
                new_path = []
                for j, value_j in enumerate(value_i):
                    new_x = value_i[j][0]
                    new_y = value_i[j][1]
                    new_m = value_i[-1 * (j + 1)][3]
                    new_z = calculate_parabola_y(new_m, factor, max_m)
                    new_path.append([new_x, new_y, new_z, new_m])

                polyline_json['paths'][i] = new_path

        # Z値の更新
        for i, value_i in enumerate(polyline_json['paths']):
            for j, value_j in enumerate(polyline_json['paths'][i]):
                polyline_json['paths'][i][j][2] = calculate_parabola_y(value_j[3], factor, max_m)
        
        row[0] = json.dumps(polyline_json)
        update_cursor.updateRow(row)

    # オブジェクト参照の解放
    del row, update_cursor

    # フィールドの削除
    arcpy.management.DeleteField(output_featureclass,temp_field_name)

    # テンポラリ フィーチャクラスの削除
    arcpy.management.Delete(temp_featureclass)

    arcpy.AddMessage('処理が完了しました。')

except arcpy.ExecuteError as e:
  arcpy.AddError('エラーが発生しました。')
  arcpy.AddError(str(e))